@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015-2024 NVIDIA Corporation
+    Copyright (c) 2015-2025 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -58,7 +58,7 @@
 #include "uvm_va_block_types.h"
 #include "uvm_linux.h"
 #include "uvm_types.h"
-#include "nv_uvm_types.h"
+#include "nv_uvm_user_types.h"
 #if UVM_IS_CONFIG_HMM() || defined(CONFIG_PCI_P2PDMA)
 #include <linux/memremap.h>
 #endif
@@ -97,18 +97,9 @@ typedef enum
 {
     // Memory type for backing user pages. On Pascal+ it can be evicted.
     UVM_PMM_GPU_MEMORY_TYPE_USER,
-    // When the Confidential Computing feature is enabled, the protected flavor
-    // allocates memory out of the VPR region. When it's disabled, all flavors
-    // have no effects and are equivalent to the base type.
-    UVM_PMM_GPU_MEMORY_TYPE_USER_PROTECTED = UVM_PMM_GPU_MEMORY_TYPE_USER,
-    UVM_PMM_GPU_MEMORY_TYPE_USER_UNPROTECTED,
 
     // Memory type for internal UVM allocations. It cannot be evicted.
     UVM_PMM_GPU_MEMORY_TYPE_KERNEL,
-    // See user types for the behavior description when the Confidential
-    // Computing feature is ON or OFF.
-    UVM_PMM_GPU_MEMORY_TYPE_KERNEL_PROTECTED = UVM_PMM_GPU_MEMORY_TYPE_KERNEL,
-    UVM_PMM_GPU_MEMORY_TYPE_KERNEL_UNPROTECTED,
 
     // Number of types - MUST BE LAST.
     UVM_PMM_GPU_MEMORY_TYPE_COUNT
@@ -116,17 +107,6 @@ typedef enum
 
 const char *uvm_pmm_gpu_memory_type_string(uvm_pmm_gpu_memory_type_t type);
 
-// Returns true if the given memory type is used to back user pages.
-bool uvm_pmm_gpu_memory_type_is_user(uvm_pmm_gpu_memory_type_t type);
-
-// Returns true if the given memory type is used to back internal UVM
-// allocations.
-static bool uvm_pmm_gpu_memory_type_is_kernel(uvm_pmm_gpu_memory_type_t type)
-{
-    return !uvm_pmm_gpu_memory_type_is_user(type);
-}
-
-bool uvm_pmm_gpu_memory_type_is_protected(uvm_pmm_gpu_memory_type_t type);
 
 typedef enum
 {
@@ -171,7 +151,7 @@ typedef enum
     // VA block lock.
     UVM_PMM_ALLOC_FLAGS_EVICT = (1 << 0),
 
-    // Do not use batching in this call if PMA page allocaion is required
+    // Do not use batching in this call if PMA page allocation is required
     UVM_PMM_ALLOC_FLAGS_DONT_BATCH = (1 << 1),
 
     UVM_PMM_ALLOC_FLAGS_MASK = (1 << 2) - 1
@@ -210,6 +190,10 @@ typedef uvm_chunk_size_t uvm_chunk_sizes_mask_t;
 
 typedef struct uvm_pmm_gpu_chunk_suballoc_struct uvm_pmm_gpu_chunk_suballoc_t;
 
+static inline void uvm_pmm_devmem_exit(void)
+{
+}
+
 static inline void uvm_pmm_gpu_device_p2p_init(uvm_gpu_t *gpu) {}
 static inline void uvm_pmm_gpu_device_p2p_deinit(uvm_gpu_t *gpu) {}
 
@@ -245,6 +229,11 @@ struct uvm_gpu_chunk_struct
 
         // This flag indicates an allocated user chunk is referenced by a device
         // private struct page PTE and therefore expects a page_free() callback.
+        // The flag is only for sanity checking since uvm_pmm_gpu_free()
+        // shouldn't be called if Linux has a device private reference to this
+        // chunk and devmem_page_free() should only be called from the Linux
+        // callback if a reference was created.
+        // See uvm_hmm_va_block_service_locked() and fill_dst_pfn() for details.
         //
         // This field is always false in kernel chunks.
         bool is_referenced : 1;
@@ -274,6 +263,9 @@ struct uvm_gpu_chunk_struct
     // The VA block using the chunk, if any.
     // User chunks that are not backed by a VA block are considered to be
     // temporarily pinned and cannot be evicted.
+    // Note that the chunk state is normally UVM_PMM_GPU_CHUNK_STATE_ALLOCATED
+    // but can also be UVM_PMM_GPU_CHUNK_STATE_TEMP_PINNED if an HMM va_block
+    // and device private struct page have a pointer to this chunk.
     //
     // This field is always NULL in kernel chunks.
     uvm_va_block_t *va_block;
@@ -332,6 +324,13 @@ typedef struct uvm_pmm_gpu_struct
         // uvm_pmm_gpu_mark_root_chunk_(un)used().
         struct list_head va_block_unused;
 
+        // List of discarded root GPU chunks, which are still mapped on the GPU.
+        // Chunks on this list are evicted with a lower priority than unused chunks.
+        //
+        // Updated by the VA block code with
+        // uvm_pmm_gpu_mark_root_chunk_discarded().
+        struct list_head va_block_discarded;
+
         // List of root chunks used by VA blocks
         struct list_head va_block_used;
 
@@ -389,10 +388,12 @@ static void uvm_gpu_chunk_set_size(uvm_gpu_chunk_t *chunk, uvm_chunk_size_t size
 // use it if the owning GPU is retained.
 uvm_gpu_t *uvm_gpu_chunk_get_gpu(const uvm_gpu_chunk_t *chunk);
 
-// Returns true if the memory type of the chunk is a user type.
+// Returns true for user chunks.
 static bool uvm_gpu_chunk_is_user(const uvm_gpu_chunk_t *chunk)
 {
-    return uvm_pmm_gpu_memory_type_is_user(chunk->type);
+    UVM_ASSERT(chunk->type < UVM_PMM_GPU_MEMORY_TYPE_COUNT);
+
+    return chunk->type == UVM_PMM_GPU_MEMORY_TYPE_USER;
 }
 
 // Return the first struct page corresponding to the physical address range
@@ -404,22 +405,25 @@ static bool uvm_gpu_chunk_is_user(const uvm_gpu_chunk_t *chunk)
 // page containing the chunk's starting address.
 struct page *uvm_gpu_chunk_to_page(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);
 
+// Return the physical address of the given chunk. The GPU must support
+// coherence, (uvm_parent_gpu_is_coherent() should return true).
+NvU64 uvm_gpu_chunk_to_sys_addr(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);
+
 // User memory allocator.
 //
 // Allocates num_chunks chunks of size chunk_size in caller-supplied array
 // (chunks).
 //
-// Returned chunks are in the TEMP_PINNED state, requiring a call to either
-// uvm_pmm_gpu_unpin_allocated, uvm_pmm_gpu_unpin_referenced, or
-// uvm_pmm_gpu_free. If a tracker is passed in, all
-// the pending operations on the allocated chunks will be added to it
+// Returned chunks are in the TEMP_PINNED state, requiring a call to
+// uvm_pmm_gpu_unpin_allocated or uvm_pmm_gpu_free. If a tracker is passed in,
+// all the pending operations on the allocated chunks will be added to it
 // guaranteeing that all the entries come from the same GPU as the PMM.
 // Otherwise, when tracker is NULL, all the pending operations will be
 // synchronized before returning to the caller.
 //
 // Each of the allocated chunks list nodes (uvm_gpu_chunk_t::list) can be used
-// by the caller until the chunk is unpinned (uvm_pmm_gpu_unpin_allocated,
-// uvm_pmm_gpu_unpin_referenced) or freed (uvm_pmm_gpu_free). If used, the list
+// by the caller until the chunk is unpinned (uvm_pmm_gpu_unpin_allocated)
+// or freed (uvm_pmm_gpu_free). If used, the list
 // node has to be returned to a valid state before calling either of the APIs.
 //
 // In case of an error, the chunks array is guaranteed to be cleared.
@@ -452,12 +456,6 @@ NV_STATUS uvm_pmm_gpu_alloc_kernel(uvm_pmm_gpu_t *pmm,
 // Can only be used on user memory.
 void uvm_pmm_gpu_unpin_allocated(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk, uvm_va_block_t *va_block);
 
-// Unpin a temporarily pinned chunk, set its reverse map to a VA block, and
-// mark it as referenced.
-//
-// Can only be used on user memory.
-void uvm_pmm_gpu_unpin_referenced(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk, uvm_va_block_t *va_block);
-
 // Free a user or kernel chunk. Temporarily pinned chunks are unpinned.
 //
 // The tracker is optional and a NULL tracker indicates that no new operation
@@ -525,22 +523,6 @@ void uvm_pmm_gpu_sync(uvm_pmm_gpu_t *pmm);
 // Mark an allocated chunk as evicted
 void uvm_pmm_gpu_mark_chunk_evicted(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);
 
-// Returns the physical address for use by accessing_gpu of a vidmem allocation
-// on the peer pmm->gpu. This address can be used for making PTEs on
-// accessing_gpu, but not for copying between the two GPUs. For that, use
-// uvm_gpu_peer_copy_address.
-uvm_gpu_phys_address_t uvm_pmm_gpu_peer_phys_address(uvm_pmm_gpu_t *pmm,
-                                                     uvm_gpu_chunk_t *chunk,
-                                                     uvm_gpu_t *accessing_gpu);
-
-// Returns the physical or virtual address for use by accessing_gpu to copy to/
-// from a vidmem allocation on the peer pmm->gpu. This may be different from
-// uvm_gpu_peer_phys_address to handle CE limitations in addressing peer
-// physical memory directly.
-uvm_gpu_address_t uvm_pmm_gpu_peer_copy_address(uvm_pmm_gpu_t *pmm,
-                                                uvm_gpu_chunk_t *chunk,
-                                                uvm_gpu_t *accessing_gpu);
-
 // Mark a user chunk as used
 //
 // If the chunk is pinned or selected for eviction, this won't do anything. The
@@ -551,6 +533,9 @@ void uvm_pmm_gpu_mark_root_chunk_used(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk
 // Mark an allocated user chunk as unused
 void uvm_pmm_gpu_mark_root_chunk_unused(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);
 
+// Mark an allocated chunk as discarded
+void uvm_pmm_gpu_mark_root_chunk_discarded(uvm_pmm_gpu_t *pmm, uvm_gpu_chunk_t *chunk);
+
 static bool uvm_gpu_chunk_same_root(uvm_gpu_chunk_t *chunk1, uvm_gpu_chunk_t *chunk2)
 {
     return UVM_ALIGN_DOWN(chunk1->address, UVM_CHUNK_SIZE_MAX) == UVM_ALIGN_DOWN(chunk2->address, UVM_CHUNK_SIZE_MAX);
@@ -607,6 +592,10 @@ static uvm_chunk_size_t uvm_chunk_find_prev_size(uvm_chunk_sizes_mask_t chunk_si
 // retained, and it's up to the caller to release them.
 NvU32 uvm_pmm_gpu_phys_to_virt(uvm_pmm_gpu_t *pmm, NvU64 phys_addr, NvU64 region_size, uvm_reverse_map_t *out_mappings);
 
+// Allocate and initialise struct page data in the kernel to support HMM.
+NV_STATUS uvm_pmm_devmem_init(uvm_parent_gpu_t *gpu);
+void uvm_pmm_devmem_deinit(uvm_parent_gpu_t *parent_gpu);
+
 // Iterates over every size in the input mask from smallest to largest
 #define for_each_chunk_size(__size, __chunk_sizes)                                  \
     for ((__size) = (__chunk_sizes) ? uvm_chunk_find_first_size(__chunk_sizes) :    \
